package model

import (
	"database/sql"
	"log"
	"time"

	"stevenlr.com/locker/utils"
)

func GenerateLockToken() (string, error) {
	return utils.GenerateRandomString(66)
}

type Lock struct {
	Id        UUID
	Name      string
	StartTime Time
	EndTime   Time
	Owner     UUID
	Token     string
}

func (self Lock) IsFinished() bool {
	return MakeTimeNow().Compare(self.EndTime) >= 0
}

func InsertLock(tx *sql.Tx, name string, seconds int, ownerId UUID) error {
	now := MakeTimeNow()
	end := Time(time.Time(now).Add(time.Duration(seconds) * time.Second))
	id := MakeUUID()
	token, _ := GenerateLockToken()
	_, err := tx.Exec(`
		INSERT INTO Lock VALUES ($1, $2, $3, $4, $5, $6)`, id, name, now, end, ownerId, token)
	return err
}

func GetLocksForUser(db *sql.DB, owner UUID) []Lock {
	rows, err := db.Query("SELECT Id, Name FROM Lock WHERE Owner=$1", owner)
	if err != nil {
		log.Fatalln(err)
	}

	locks := []Lock{}
	for rows.Next() {
		var t Lock
		if err := rows.Scan(&t.Id, &t.Name); err == nil {
			locks = append(locks, t)
		}
	}

	return locks
}

func GetLockForUser(db *sql.DB, id UUID, userId UUID) *Lock {
	row := db.QueryRow("SELECT Id, Name, StartTime, EndTime, Owner, Token FROM Lock WHERE Id=$1 AND Owner=$2", id, userId)

	var t Lock
	if err := row.Scan(&t.Id, &t.Name, &t.StartTime, &t.EndTime, &t.Owner, &t.Token); err == nil {
		return &t
	}

	return nil
}

func GetLockWithToken(db *sql.DB, id UUID, token string) *Lock {
	row := db.QueryRow("SELECT Id, Name, StartTime, EndTime, Owner, Token FROM Lock WHERE Id=$1 AND Token=$2", id, token)

	var t Lock
	if err := row.Scan(&t.Id, &t.Name, &t.StartTime, &t.EndTime, &t.Owner, &t.Token); err == nil {
		return &t
	}

	return nil
}

func DeleteLock(db *sql.DB, id UUID, userId UUID) bool {
	res, err := db.Exec("DELETE FROM Lock WHERE Id=$1 AND Owner=$2", id, userId)
	if err != nil {
		return false
	}

	affected, err := res.RowsAffected()
	return err == nil && affected == 1
}

func UpdateLockEndTime(db *sql.DB, id UUID, endTime Time, userId UUID) bool {
	res, err := db.Exec("UPDATE Lock SET EndTime=$1 WHERE Id=$2 AND Owner=$3", endTime, id, userId)
	if err != nil {
		return false
	}

	affected, err := res.RowsAffected()
	return err == nil && affected == 1
}

func RegenerateLockToken(db *sql.DB, id UUID, userId UUID) bool {
	newToken, err := GenerateLockToken()
	if err != nil {
		return false
	}

	res, err := db.Exec("UPDATE Lock SET Token=$1 WHERE Id=$2 AND Owner=$3", newToken, id, userId)
	if err != nil {
		return false
	}

	affected, err := res.RowsAffected()
	return err == nil && affected == 1
}
